@@ -12,7 +12,7 @@ from argparse import (
     Namespace,
 )
 
-from simpleconf import Config
+from simpleconf import Config  # type: ignore[import-untyped]
 
 from . import type_
 from .utils import (
@@ -40,12 +40,13 @@ from .action import (
 from .formatter import ChargedHelpFormatter
 
 if TYPE_CHECKING:
-    from argparse import _ActionT, _FormatterClass
+    from argparse import Action, _FormatterClass
 
 
 @add_attribute("level", 0)
 class ArgumentParser(APArgumentParser):
-    """Class to handle command line arguments and configuration files"""
+    """Supercharged ArgumentParser for parsing command line strings into
+    Python objects."""
 
     def __init__(
         self,
@@ -59,13 +60,44 @@ class ArgumentParser(APArgumentParser):
         fromfile_prefix_chars: str | None = None,
         argument_default: Any = None,
         conflict_handler: str = "error",
-        add_help: bool | str = True,
+        add_help: bool | str | Sequence[str] = True,
         allow_abbrev: bool = True,
         exit_on_error: bool = True,
         exit_on_void: bool = False,
-        pre_parse: Callable[[ArgumentParser], None] | None = None,
+        pre_parse: (
+            Callable[[ArgumentParser, Sequence[str], Namespace], None] | None
+        ) = None,
     ) -> None:
+        """Create an ArgumentParser
+
+        Args:
+            prog (str, optional): The program name
+            usage (str, optional): The usage message
+            description (str, optional): The description message
+            epilog (str, optional): The epilog message
+            parents (Sequence[ArgumentParser], optional): The parent parsers
+            formatter_class (type, optional): The formatter class
+            prefix_chars (str, optional): The prefix characters
+            fromfile_prefix_chars (str, optional): The prefix characters for
+                configuration files
+            argument_default (Any, optional): The default value for arguments
+            conflict_handler (str, optional): The conflict handler
+            add_help (bool | str | Sequence[str], optional): Whether to
+                add the help option.
+                If True (default), same as "h,help".
+                If any option ends with "+", it will add the help option with the
+                ability to show more options (e.g. "h+,help+").
+                If False, it will not add the help option
+            allow_abbrev (bool, optional): Whether to allow abbreviation
+            exit_on_error (bool, optional): Whether to exit on error
+            exit_on_void (bool, optional): Whether to exit on void arguments
+                Added by `argx`.
+            pre_parse (Callable[[ArgumentParser], None], optional): The
+                function to call before parsing.
+                Added by `argx`.
+        """
         old_add_help = add_help
+        # disable add_help to add it later
         add_help = False
         kwargs = {
             "prog": prog,
@@ -87,7 +119,7 @@ class ArgumentParser(APArgumentParser):
 
         self.exit_on_void = exit_on_void
         self.pre_parse = pre_parse
-        self._subparsers_action = None
+        self._subparsers_action: _SubParsersAction | None = None
 
         # Register our actions to override argparse's or add new ones
         self.register("action", None, StoreAction)
@@ -108,32 +140,39 @@ class ArgumentParser(APArgumentParser):
         self.register("type", "py", type_.py)
         self.register("type", "json", type_.json)
         self.register("type", "path", Path)
+        self.register("type", "anypath", type_.anypath)
         self.register("type", "auto", type_.auto)
 
         # Add help option to support + for more options
-        default_prefix = (
-            "-" if "-" in self.prefix_chars else self.prefix_chars[0]
-        )
-        if old_add_help is True:
+        default_prefix = "-" if "-" in self.prefix_chars else self.prefix_chars[0]
+        if old_add_help is not False:
+            if old_add_help is True:
+                old_add_help = "h,help"
+
+            if isinstance(old_add_help, str):
+                old_add_help = old_add_help.split(",")
+
+            old_add_help = [x.strip() for x in old_add_help]
+            help_msg = "show this help message and exit"
+            if any(x.endswith("+") for x in old_add_help):
+                help_msg = f"{help_msg} (with + for more options)"
+
+            old_add_help = [
+                (
+                    f"{default_prefix * 2}{x}"
+                    if len(x.rstrip("+")) > 1
+                    else f"{default_prefix}{x}"
+                )
+                for x in old_add_help
+            ]
+
             self.add_argument(
-                f"{default_prefix}h",
-                f"{default_prefix * 2}help",
+                *old_add_help,
                 action="help",
                 default=SUPPRESS,
-                help=_("show help message and exit"),
-            )
-        elif old_add_help == "+":
-            self.add_argument(
-                f"{default_prefix}h",
-                f"{default_prefix * 2}help",
-                f"{default_prefix}h+",
-                f"{default_prefix * 2}help+",
-                action="help",
-                default=SUPPRESS,
-                help=_(
-                    "show help message (with + to show more options) and exit"
-                ),
+                help=_(help_msg),
             )
+
         # restore add_help
         self.add_help = old_add_help  # type: ignore[assignment]
 
@@ -142,11 +181,12 @@ class ArgumentParser(APArgumentParser):
             order=-1,
         )
 
-    def add_subparsers(self, order: int = 99, **kwargs):
+    def add_subparsers(self, order: int = 99, **kwargs) -> _SubParsersAction:
         """Add subparsers to the parser.
 
         Args:
             order (int): The order of the subparsers group
+                Added by `argx`.
             **kwargs: The arguments to pass to add_argument_group()
 
         Returns:
@@ -154,15 +194,15 @@ class ArgumentParser(APArgumentParser):
         """
         action = super().add_subparsers(**kwargs)
         if self._subparsers is not self._positionals:
-            self._subparsers.order = order
+            self._subparsers.order = order  # type: ignore[assignment]
 
         # Add parent to subparsers action
         # So that subparsers.add_parser add parent to the sub-parser
-        action.parent = self
+        action.parent = self  # type: ignore[assignment]
         return action
 
     def add_subparser(self, name: str, **kwargs) -> ArgumentParser | Any:
-        """Add a subparser directly.
+        """Add a subparser directly, a shortcut to add sub-commands by `argx`.
 
         Instead of
         >>> subparsers = parser.add_subparsers(...)
@@ -191,25 +231,40 @@ class ArgumentParser(APArgumentParser):
             self._subparsers_action = self.add_subparsers(
                 title=_("subcommands"),
                 required=True,
-                dest="COMMAND"
-                if self.level == 0
-                else f"COMMAND{self.level+1}",
+                dest=(
+                    "COMMAND"
+                    if self.level == 0  # type: ignore[assignment]
+                    else f"COMMAND{self.level + 1}"  # type: ignore[assignment]
+                ),
             )
         kwargs.setdefault("help", f"The {name} command")
-        return self._subparsers_action.add_parser(
-            name, level=self.level + 1, **kwargs
-        )
+        out = self._subparsers_action.add_parser(name, **kwargs)
+        out.level = self.level + 1  # type: ignore[assignment]
+        return out
 
     add_command = add_subparser
 
-    def parse_known_args(
+    def parse_known_args(  # type: ignore[override]
         self,
         args: Sequence[str] | None = None,
         namespace: Namespace | None = None,
+        fromfile_parse: bool = True,
+        fromfile_keep: bool = False,
     ) -> tuple[Namespace, list[str]]:
         """Parse known arguments.
 
-        Modify to handle exit_on_void and @file to load default values.
+        Modify to handle exit_on_void and @file to load default values by `argx`.
+
+        Args:
+            args: The arguments to parse.
+            namespace: The namespace to use.
+            fromfile_parse: Whether to parse @file.
+                Added by `argx`.
+            fromfile_keep: Whether to keep @file in the unknown arguments.
+                Note that @file.txt file will be treated as a normal argument,
+                thus, it will be parsed and not kept anyway. This means at
+                this point, @file.txt will be expanded right away.
+                Added by `argx`.
         """
         if args is None:  # pragma: no cover
             # args default to the system args
@@ -227,26 +282,43 @@ class ArgumentParser(APArgumentParser):
             if new_args is not None:
                 args = new_args
 
-        if not self.fromfile_prefix_chars:
-            new_args = args
-        else:
-            # Setup the defaults if a configuration file is given by @config.ini
-            new_args = []
-            for arg in args:
-                if (
-                    not arg
-                    or arg[0] not in self.fromfile_prefix_chars
-                    or arg.endswith(".txt")
-                ):
-                    new_args.append(arg)
-                else:
-                    try:
-                        conf = arg[1:]
-                        if conf.endswith(".py"):
-                            conf = import_pyfile(conf)
-                        self.set_defaults_from_configs(conf)
-                    except Exception as e:
-                        self.error(f"Cannot import [{conf}]: {e}")
+        # @files to keep in the unknown arguments
+        files = []
+        # arguments passed to super().parse_known_args() for parsing
+        new_args = []
+
+        for arg in args:
+            # no fromfile_prefix_chars or normal argument
+            if (
+                not self.fromfile_prefix_chars
+                or not arg
+                or arg[0] not in self.fromfile_prefix_chars
+            ):
+                new_args.append(arg)
+
+            # fromfile_prefix_chars is set
+            # and arg is not empty
+            # and arg starts with fromfile_prefix_chars
+
+            # @file.txt is special, send it to super().parse_known_args()
+            # for parsing
+            elif arg.endswith(".txt"):
+                new_args.append(arg)
+            # @file.py, @file.json, ...
+            else:
+                if fromfile_keep:
+                    # keep @file in the unknown arguments
+                    files.append(arg)
+
+                if fromfile_parse:
+                    # parse @file to set the default values
+                    conf = arg[1:]
+                    if conf.endswith(".py"):
+                        try:
+                            conf: dict | str = import_pyfile(conf)
+                        except Exception as e:
+                            self.error(f"Cannot import [{conf}]: {e}")
+                    self.set_defaults_from_configs(conf)
 
         # add any action defaults that aren't present
         # Do this mainly for namespace actions, like "--group.abc"
@@ -260,6 +332,7 @@ class ArgumentParser(APArgumentParser):
                 setattr(ns, last_key, action.default)
 
         parsed_args, argv = super().parse_known_args(new_args, namespace)
+        argv = files + argv
         if not argv and not args and self.exit_on_void:
             self.error("No arguments provided")
 
@@ -304,22 +377,22 @@ class ArgumentParser(APArgumentParser):
                             optionalize=optionalize,
                         )
 
-    def _add_action(self, action: _ActionT) -> _ActionT:
+    def _add_action(self, action):
         """Add an action to the parser.
 
         Modify to handle namespace actions, like "--group.abc"
         """
-        if (
-            isinstance(action, APArgumentGroup)
-            or (
-                not isinstance(action, NamespaceAction)
-                and "." not in action.dest
-            )
+        if isinstance(action, APArgumentGroup) or (
+            not isinstance(action, NamespaceAction) and "." not in action.dest
         ):
-            if action.required:
-                return self._required_actions._add_action(action)
-            return super()._add_action(action)
+            if action.required:  # type: ignore[union-attr]
+                return self._required_actions._add_action(
+                    action  # type: ignore[return-value]
+                )
+            return super()._add_action(action)  # type: ignore[return-value]
 
+        # Do not transform the keys for namespace action
+        action.dest = action.option_strings[0].lstrip(self.prefix_chars)
         # Split the destination into a list of keys
         keys = action.dest.split(".")
         seq = range(len(keys) - 1, 0, -1)
@@ -334,7 +407,7 @@ class ArgumentParser(APArgumentParser):
             for action_group in self._action_groups:
                 if (
                     isinstance(action_group, _NamespaceArgumentGroup)
-                    and action_group.name == ns_key
+                    and action_group.name == ns_key  # type: ignore[union-attr]
                 ):
                     group = action_group
                     break
@@ -347,6 +420,33 @@ class ArgumentParser(APArgumentParser):
 
         return group._add_action(action)
 
+    def get_action(
+        self,
+        dest: str,
+        include_ns_group: bool = False,
+    ) -> Action | _NamespaceArgumentGroup | None:
+        """Get an action by its destination.
+
+        Added by `argx`.
+
+        Args:
+            dest: The destination of the action
+
+        Returns:
+            Action: The action. None if not found.
+        """
+        if include_ns_group:
+            for action_group in self._action_groups:
+                if isinstance(action_group, _NamespaceArgumentGroup):
+                    if action_group.name == dest:  # type: ignore[union-attr]
+                        return action_group
+
+        for action in self._actions:
+            if action.dest == dest:
+                return action
+
+        return None
+
     def add_namespace(
         self,
         name: str,
@@ -368,14 +468,15 @@ class ArgumentParser(APArgumentParser):
         for action_group in self._action_groups:
             if (
                 isinstance(action_group, _NamespaceArgumentGroup)
-                and action_group.name == name
+                and action_group.name == name  # type: ignore[union-attr]
             ):
                 raise ValueError(f"Namespace '{name}' already exists")
 
         if title is None:
             title = f"{_('namespace')} <{name}>"
 
-        group = _NamespaceArgumentGroup(self, title, name=name, **kwargs)
+        group = _NamespaceArgumentGroup(self, title, **kwargs)
+        group.name = name  # type: ignore[assignment]
         self._action_groups.append(group)
         return group
 
@@ -410,16 +511,14 @@ class ArgumentParser(APArgumentParser):
         formatter = self._get_formatter()
 
         # usage
-        formatter.add_usage(
-            self.usage, self._actions, self._mutually_exclusive_groups
-        )
+        formatter.add_usage(self.usage, self._actions, self._mutually_exclusive_groups)
 
         # description
         formatter.add_text(self.description)
         # positionals, optionals and user-defined groups
         for action_group in sorted(
             self._action_groups,
-            key=lambda x: (x.order, x.title),
+            key=lambda x: (x.order, x.title),  # type: ignore[return-value]
         ):
             if not plus and not showable(action_group):
                 for action in action_group._group_actions:
@@ -429,13 +528,13 @@ class ArgumentParser(APArgumentParser):
 
             formatter.start_section(
                 "\033[1m\033[4m"
-                f"{format_title(action_group.title)}"
+                f"{format_title(action_group.title)}"  # type: ignore[arg-type]
                 "\033[0m\033[0m"
             )
             formatter.add_text(action_group.description)
             formatter.add_arguments(  # type: ignore[call-arg]
                 action_group._group_actions,
-                plus,
+                plus,  # type: ignore[arg-type]
             )
             formatter.end_section()
 
@@ -507,22 +606,22 @@ class ArgumentParser(APArgumentParser):
         default: Any = None,
     ) -> Any:
         # Allow type to be string
-        if (
-            registry_name == "type"
-            and isinstance(value, str)
-            and default == value
-        ):
+        if registry_name == "type" and isinstance(value, str) and default == value:
             import builtins
 
             # "int", "float", "str", "open", etc
-            return self._registries[registry_name].get(
+            got = self._registries[registry_name].get(
                 value,
                 getattr(builtins, value, None),
             )
+            if not callable(got):
+                raise ValueError(f"Invalid type '{value}'")
+
+            return got
 
         # Not using super() because _ArgumentGroup can also use it
         return APArgumentGroup._registry_get(
-            self,
+            self,  # type: ignore[return-value]
             registry_name,
             value,
             default,
@@ -530,7 +629,9 @@ class ArgumentParser(APArgumentParser):
 
     @classmethod
     def from_configs(cls, *configs: dict | str, **kwargs) -> ArgumentParser:
-        """Create an ArgumentParser from a configuration file
+        """Create an ArgumentParser from a configuration file.
+
+        Added by `argx`.
 
         Args:
             *configs: The configuration files or dicts to load
@@ -569,4 +670,4 @@ class _ArgumentGroup(APArgumentGroup):
 
 @add_attribute("name")
 class _NamespaceArgumentGroup(_ArgumentGroup):
-    ...
+    pass
